//=============================================================================
//
//  CLASS LightMap - IMPLEMENTATION
//
//=============================================================================

//== INCLUDES =================================================================

#include "LightMap.hh"


//== IMPLEMENTATION ========================================================== 


LightMap::
LightMap(GLenum texunit, GLint _mode)
  : TextureCubeMap(texunit), mode_(_mode)
{
}


//-----------------------------------------------------------------------------


#define CUBE_MAP_SIZE 256


void
LightMap::
update()
{
  int components = 0;
  if (mode_ == GL_RGBA)
  {
    // use four components to store colored light
    components = 4;
  }
  else
  {
    // use 2 components to store achromatic light separted in specular
    // and non-specular component
    components = 2;
  }

//  GLubyte CubeMap[6][CUBE_MAP_SIZE][CUBE_MAP_SIZE][components];
  static GLubyte CubeMap[6][CUBE_MAP_SIZE][CUBE_MAP_SIZE][4];

  OpenMesh::Vec3f  n;
  OpenMesh::Vec4uc c;


  // update OpenGL material & light parameters
  get_GL_params();


  // calc 6 sides
  enum {CUBE_POS_X=0, CUBE_NEG_X=1,
	CUBE_POS_Y=2, CUBE_NEG_Y=3,
	CUBE_POS_Z=4, CUBE_NEG_Z=5};

  GLenum faces[6] = {
    GL_TEXTURE_CUBE_MAP_POSITIVE_X,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
    GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
  };

  int i, j, k;


  for (i=0; i<CUBE_MAP_SIZE; ++i)
  {
    float t = 1.0 / (2 * CUBE_MAP_SIZE) + (float) i / CUBE_MAP_SIZE;
    t = 2.0 * t - 1.0;
    for (j=0; j<CUBE_MAP_SIZE; ++j)
    {
      float s = 1.0 / (2 * CUBE_MAP_SIZE) + (float) j / CUBE_MAP_SIZE;
      s = 2.0 * s - 1.0;

      n[0] = 1;
      n[1] = -t;
      n[2] = -s;
      c    = light(n.normalize());
      for (k=0; k<components; ++k) CubeMap[CUBE_POS_X][i][j][k] = c[k];

      n[0] = -1;
      n[1] = -t;
      n[2] = s;
      c    = light(n.normalize());
      for (k=0; k<components; ++k) CubeMap[CUBE_NEG_X][i][j][k] = c[k];


      n[1] = 1;
      n[0] = s;
      n[2] = t;
      c    = light(n.normalize());
      for (k=0; k<components; ++k) CubeMap[CUBE_POS_Y][i][j][k] = c[k];

      n[1] = -1;
      n[0] = s;
      n[2] = -t;
      c    = light(n.normalize());
      for (k=0; k<components; ++k) CubeMap[CUBE_NEG_Y][i][j][k] = c[k];


      n[2] = 1;
      n[0] = s;
      n[1] = -t;
      c    = light(n.normalize());
      for (k=0; k<components; ++k) CubeMap[CUBE_POS_Z][i][j][k] = c[k];

      n[2] = -1;
      n[0] = -s;
      n[1] = -t;
      c    = light(n.normalize());
      for (k=0; k<components; ++k) CubeMap[CUBE_NEG_Z][i][j][k] = c[k];
    }
  }



  // copy images to GL
  // internal format is determined by mode_; 
  // currently GL_RGBA, GL_LUMINANCE_ALPHA
  bind();
  for (i=0; i<6; ++i)
    glTexImage2D(faces[i],
		 0,                 // level
		 mode_,             // internal format
		 CUBE_MAP_SIZE,     // width
		 CUBE_MAP_SIZE,     // height
		 0,                 // border
		 mode_,             // format
		 GL_UNSIGNED_BYTE,  // type
		 CubeMap[i]);       // pixel data


  // set texture parameters
  parameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  parameter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  parameter(GL_TEXTURE_WRAP_R,     GL_CLAMP_TO_EDGE);
  parameter(GL_TEXTURE_WRAP_S,     GL_CLAMP_TO_EDGE);
  parameter(GL_TEXTURE_WRAP_T,     GL_CLAMP_TO_EDGE);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}


//-----------------------------------------------------------------------------


void
LightMap::
get_GL_params()
{
  // get material properties

  glGetMaterialfv(GL_FRONT, GL_EMISSION,   mat_emission_);
  glGetMaterialfv(GL_FRONT, GL_AMBIENT,    mat_ambient_);
  glGetMaterialfv(GL_FRONT, GL_DIFFUSE,    mat_diffuse_);
  glGetMaterialfv(GL_FRONT, GL_SPECULAR,   mat_specular_);
  glGetMaterialfv(GL_FRONT, GL_SHININESS,  &mat_shininess_);




  // get light sources

  glGetFloatv(GL_LIGHT_MODEL_AMBIENT, ambient_light_);

  GLuint  names[8] = { GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3,
		       GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7 };

  for (unsigned int i=0; i<8; ++i)
  {
    glGetLightfv(names[i], GL_POSITION,  lights_[i].position);
    glGetLightfv(names[i], GL_AMBIENT,   lights_[i].ambient);
    glGetLightfv(names[i], GL_DIFFUSE,   lights_[i].diffuse);
    glGetLightfv(names[i], GL_SPECULAR,  lights_[i].specular);
    lights_[i].enabled = glIsEnabled(names[i]);
  }
}


//-----------------------------------------------------------------------------


OpenMesh::Vec4uc
LightMap::
light(const Normal& _n) const
{
  Color       col(0,0,0,0);
  Color       col2(0,0,0,0);
  OpenMesh::Vec3f  V(0,0,1), L, H;
  float       diff_dot, spec_dot;


  // self emission
  col += mat_emission_;


  // global ambient light
  col += ambient_light_ * mat_ambient_;


  // contribution from all (enabled) light sources
  for (int i=0; i<8; ++i)
  {
    if(lights_[i].enabled)
    {
      if (lights_[i].position[3] != 0.0) 
      {
	std::cerr << "LightMap: light " << i << " is not a directional "
		  << "light source!\n";
	continue;
      }


      // light position/direction
      L[0] = lights_[i].position[0];
      L[1] = lights_[i].position[1];
      L[2] = lights_[i].position[2];
      L.normalize();


      // halfway vector (Blinn lighting)
      H = (L + V).normalize();


      // ambient
      col += lights_[i].ambient * mat_ambient_;


      // diffuse
      diff_dot = (L | _n);
      if (diff_dot > 0.0)
      {
	col += diff_dot * lights_[i].diffuse * mat_diffuse_;


	// specular
	spec_dot = (H | _n);
	if (spec_dot > 0.0)
	{
	  col2 += ((float)pow(spec_dot, mat_shininess_) *
		   lights_[i].specular * mat_specular_);
	}
      }
    }
  }

  if (mode_ == GL_RGBA)
  {
    col += col2;
  }
  else
  {
    // just two values for achromatic light with separate specular component
    // -> copy specular component to 2nd field of col
    col[1] = col2[0];
  }

  // clamp to [0,1]
  col.maximize(Color(0,0,0,0));
  col.minimize(Color(1,1,1,1));

  return OpenMesh::Vec4uc((unsigned char)(col[0]*255.0),
			  (unsigned char)(col[1]*255.0),
			  (unsigned char)(col[2]*255.0),
			  (unsigned char)(col[3]*255.0));
}


//=============================================================================
