/******************************************************************************
                       C++ Implementation: TSMVisualVBO
                            -------------------
    description       :
    begin             : Wed Mar 23 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "TSMVisualVBO.h"

#ifdef WIN32
  #include <windows.h>
#endif
#include <GL/glew.h>
#include "defs.h"

#include "TwoScaleModel.h"

namespace
{

IVisualization* create_VBO_vis(const TwoScaleModel& _tsm, int /*argc*/, char** /*argv*/)
{ return new TSMVisualVBO(_tsm); }

struct RegisterVBOVisCreator
{
  RegisterVBOVisCreator()
  { VisualizationFactory::connect('1', create_VBO_vis); }
};

RegisterVBOVisCreator register_vbo_vis_creator__;

}

#ifdef PRECOMPUTE_STATIC_FRAC  
struct FSVBlockHeader
{
  float pt_displ_;
  Vec3  t0_, t1_;
  Vec3  prec_pt_, prec_du_, prec_dv_;
};
#endif

TSMVisualVBO::TSMVisualVBO(const TwoScaleModel& _tsm)
: tsm_(_tsm)
{ 
#ifdef PRECOMPUTE_STATIC_FRAC  
  uint n_fs_vertices = tsm_.fs_vertices_.size();
  
  std::vector<Vec3> prec_pt(n_fs_vertices), prec_dv(n_fs_vertices), prec_du(n_fs_vertices);
  CVCoeffArray cv_coeffs;
  
  uint n_modified_cvs = tsm_.calc_geometry_split(&prec_pt.front(), &prec_du.front(), &prec_dv.front(), cv_coeffs);
  
  for (uint i = 0, n = cv_coeffs.size(); i < n; ++i)
  {//find out which control vertices are modified
    if (!cv_coeffs[i].empty())
    {
      modified_cvs_.push_back(i);    
    }
  }
  CHECK(n_modified_cvs == modified_cvs_.size());//we got everything, right?
  uint fsv_block_size = sizeof(FSVBlockHeader) + n_modified_cvs*sizeof(Vec3);
  uint fsv_blocks_mem = fsv_block_size*n_fs_vertices;
  printf("Memory required for the fine scale vertices representation: %.4f Mb.\n", fsv_blocks_mem/(1024.0*1024.0));
  fsv_blocks_ = new char[fsv_blocks_mem];  
  
  char* fsv_block_it = fsv_blocks_;
/*  uint fsv_block_n_floats = fsv_block_size/sizeof(float);
  CHECK((fsv_block_size % sizeof(float)) == 0);//should contain only floats*/
  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    //first fill up the header 
    FSVBlockHeader& fsv_block_header = reinterpret_cast<FSVBlockHeader&>(*fsv_block_it);
    const FineScaleVertex& fsv = tsm_.fs_vertices_[i];
    fsv_block_header.pt_displ_ = fsv.pt_displ_;
    fsv_block_header.t0_ = Vec3(fsv.t0_displ_n_, fsv.t0_displ_du_, fsv.t0_displ_dv_);
    fsv_block_header.t1_ = Vec3(fsv.t1_displ_n_, fsv.t1_displ_du_, fsv.t1_displ_dv_);
    fsv_block_header.prec_pt_ = prec_pt[i];
    fsv_block_header.prec_du_ = prec_du[i];
    fsv_block_header.prec_dv_ = prec_dv[i];
    fsv_block_it += sizeof(FSVBlockHeader);
    //now append the coefficients
    for (uint j = 0; j < n_modified_cvs; ++j)
    {
      Vec3& coeffs = reinterpret_cast<Vec3&>(*fsv_block_it);
      coeffs = cv_coeffs[modified_cvs_[j]][i];
      fsv_block_it += sizeof(Vec3);
    }
  } 
  uint fsv_geom_size = sizeof(FSVGeometry)*n_fs_vertices;
  
  OGL_CHECK(::glGenBuffersARB(1, &vbo_fsv_geom_));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_fsv_geom_));//bind normal array
  OGL_CHECK(::glBufferDataARB(GL_ARRAY_BUFFER_ARB, fsv_geom_size, NULL, GL_DYNAMIC_DRAW_ARB));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer
  fsv_geom_ = NULL;
  
#else
  //generate separate vbos for vertices & normals
  uint vertices_size = tsm_.fs_vertices_.size()*sizeof(Vec3);
  
  OGL_CHECK(::glGenBuffersARB(1, &vbo_vertices_));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));//bind vertex array
  OGL_CHECK(::glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertices_size,  NULL, GL_DYNAMIC_DRAW_ARB));
  
  OGL_CHECK(::glGenBuffersARB(1, &vbo_normals_));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_normals_));//bind normal array
  OGL_CHECK(::glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertices_size, NULL, GL_DYNAMIC_DRAW_ARB));
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer

  vertices_ = normals_ = NULL;

#endif//PRECOMPUTE_STATIC_FRAC

  //Generate indices' VBO
  OGL_CHECK(::glGenBuffersARB(1, &vbo_indices_));

  n_indices_ = tsm_.fs_tris_.size()*3;
  uint indices_size =  tsm_.fs_tris_.size()*sizeof(Triangle);
  
  OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo_indices_));//bind indices array
  OGL_CHECK(::glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, indices_size,
                              &tsm_.fs_tris_.front(), GL_STATIC_DRAW_ARB));
  OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0));//unbind indicies buffer
}

TSMVisualVBO::~TSMVisualVBO()
{
#ifdef PRECOMPUTE_STATIC_FRAC  
  delete fsv_blocks_;
  OGL_CHECK(::glDeleteBuffersARB(1, &vbo_fsv_geom_));
#else
  OGL_CHECK(::glDeleteBuffersARB(1, &vbo_vertices_));
  OGL_CHECK(::glDeleteBuffersARB(1, &vbo_normals_));  
#endif  
  OGL_CHECK(::glDeleteBuffersARB(1, &vbo_indices_));
}

bool TSMVisualVBO::map_geometry_vbos()
{
#ifdef PRECOMPUTE_STATIC_FRAC  
  CHECK(fsv_geom_ == NULL);
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_fsv_geom_));
  OGL_CHECK(fsv_geom_ = (FSVGeometry*)::glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer
  return fsv_geom_ != NULL;

#else
  CHECK(vertices_ == NULL && normals_ == NULL);
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));
  OGL_CHECK(vertices_ = (float*)::glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_normals_));
  OGL_CHECK(normals_ = (float*)::glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_WRITE_ONLY_ARB));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer
  return vertices_ != NULL && normals_ != NULL;
#endif
}

bool TSMVisualVBO::unmap_geometry_vbos()
{
#ifdef PRECOMPUTE_STATIC_FRAC  
  CHECK(fsv_geom_ != NULL);
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_fsv_geom_));
  OGL_CHECK(bool unmap_vbo_res = ::glUnmapBufferARB(GL_ARRAY_BUFFER_ARB) == 1); 
  if (!unmap_vbo_res)
  {
    printf("Warning: Unmapping vbo_fsv_geom_ failed!\n");
    OGL_CHECK(::glUnmapBufferARB(GL_ARRAY_BUFFER_ARB));//unmap second time?
  }
  fsv_geom_ = NULL;
  
  return unmap_vbo_res;
#else
  CHECK(vertices_ != NULL && normals_ != NULL);
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));
  OGL_CHECK(bool unmap_vvbo_res = ::glUnmapBufferARB(GL_ARRAY_BUFFER_ARB) == 1); 
  if (!unmap_vvbo_res)
  {
    printf("Warning: Unmapping vbo_vertices_ failed!\n");
    OGL_CHECK(::glUnmapBufferARB(GL_ARRAY_BUFFER_ARB));//unmap second time?
  }
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_normals_));
  OGL_CHECK(bool unmap_nvbo_res = ::glUnmapBufferARB(GL_ARRAY_BUFFER_ARB) == 1); 
  if (!unmap_nvbo_res)
  {
    printf("Warning: Unmapping vbo_normals_ failed!\n");
    OGL_CHECK(::glUnmapBufferARB(GL_ARRAY_BUFFER_ARB));//unmap second time?
  }
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer
  vertices_ = normals_ = NULL;
  
  return unmap_vvbo_res && unmap_nvbo_res;
#endif
}

void TSMVisualVBO::update_geometry()
{
  map_geometry_vbos();
#ifdef PRECOMPUTE_STATIC_FRAC  
  uint n_modified_cvs = modified_cvs_.size();
  std::vector<Vec3> mod_cvs(n_modified_cvs);//should be in stack memory (faster access)
  for (uint i = 0; i < n_modified_cvs; ++i)
  {//cache the modified vertices for (probably) faster access time
    mod_cvs[i] = tsm_.cm_vertices_[modified_cvs_[i]];
  }
  uint n_fs_vertices = tsm_.fs_vertices_.size();
  FSVGeometry* fsv_geom_it = fsv_geom_;
  const char* fsv_block_it = fsv_blocks_;
  for (uint i = 0; i < n_fs_vertices; ++i, ++fsv_geom_it)
  {
    const FSVBlockHeader& fsv_block_header = reinterpret_cast<const FSVBlockHeader&>(*fsv_block_it);
    Vec3 pt(fsv_block_header.prec_pt_), du(fsv_block_header.prec_du_), dv(fsv_block_header.prec_dv_); 
    fsv_block_it += sizeof(FSVBlockHeader);
    
    for (uint j = 0; j < n_modified_cvs; ++j)
    {//compute "foot" frame
      const Vec3& cv_coeffs = reinterpret_cast<const Vec3&>(*fsv_block_it);
      pt.add_linear(cv_coeffs[0], mod_cvs[j]);
      du.add_linear(cv_coeffs[1], mod_cvs[j]);
      dv.add_linear(cv_coeffs[2], mod_cvs[j]);
      fsv_block_it += sizeof(Vec3);
    }
    
    Vec3 sd_n;
    sd_n.set_cross(du, dv);
    sd_n.normalize();
    
    pt.add_linear(fsv_block_header.pt_displ_, sd_n);
    
    Vec3 t0, t1;
    t0.set_linear(fsv_block_header.t0_[0], sd_n);
    t0.add_linear(fsv_block_header.t0_[1], du);
    t0.add_linear(fsv_block_header.t0_[2], dv);
    
    t1.set_linear(fsv_block_header.t1_[0], sd_n);
    t1.add_linear(fsv_block_header.t1_[1], du);
    t1.add_linear(fsv_block_header.t1_[2], dv);
    
    fsv_geom_it->pt_ = pt;
    fsv_geom_it->n_.set_cross(t0, t1);
  } 
#else
   tsm_.calc_geometry(reinterpret_cast<Vec3*>(vertices_), reinterpret_cast<Vec3*>(normals_));
#endif

  unmap_geometry_vbos();
}

/// IVisualization API

bool TSMVisualVBO::init()
{ return true; }//this rendering path should be always valid

void TSMVisualVBO::draw(ModeEnum _mode)
{
  update_geometry();
  
  OGL_CHECK(::glEnableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glEnableClientState(GL_NORMAL_ARRAY));
#ifdef PRECOMPUTE_STATIC_FRAC  
  CHECK(fsv_geom_ == NULL);
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_fsv_geom_));
  OGL_CHECK(::glVertexPointer(3, GL_FLOAT, sizeof(FSVGeometry), (char*) NULL));
  OGL_CHECK(::glNormalPointer(GL_FLOAT, sizeof(FSVGeometry), (char*) NULL + sizeof(Vec3)));
#else
  CHECK(vertices_ == NULL && normals_ == NULL);
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));
  OGL_CHECK(::glVertexPointer(3, GL_FLOAT, 0, (char *) NULL));
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_normals_));
  OGL_CHECK(::glNormalPointer(GL_FLOAT, 0, (char *) NULL));
#endif  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer

  if (_mode == MODE_POINTS)
  {
    OGL_CHECK(::glDrawArrays(GL_POINTS, 0, tsm_.fs_vertices_.size()));
  }
  else
  {
//   OGL_CHECK(::glDrawElements(GL_TRIANGLES, tsm_.fs_tris_.size(), GL_UNSIGNED_INT, &tsm_.fs_tris_.front()));
    OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo_indices_));
    OGL_CHECK(::glDrawElements(GL_TRIANGLES, n_indices_, GL_UNSIGNED_INT, (char *) NULL));
    OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0));//unbind indicies buffer
  }

  OGL_CHECK(::glDisableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glDisableClientState(GL_NORMAL_ARRAY));
}

void TSMVisualVBO::dump()
{}


