/******************************************************************************
                       C++ Implementation: TSMVisualVP
                            -------------------
    description       :
    begin             : Wed Mar 23 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "TSMVisualVP.h"

#include <stdio.h>
#include <stdlib.h>
#include <vector>
#ifdef WIN32
  #include <windows.h>
  #define snprintf _snprintf
#endif
#include <GL/glew.h>
#include "defs.h"

#include "vbo.h"
#include "VertexProgram.h"
#include "TwoScaleModel.h"

/// TSMVisualVP

#define INTERLEAVED_COEFF

namespace
{

IVisualization* create_VP_vis(const TwoScaleModel& _tsm, int argc, char** argv)
{ 
  for (int i=1; i < argc; ++i)
  {
    if (strcmp(argv[i], "-vsa") == 0 && (i + 1) < argc)
    {//set "vertex program for atributes" file name
      return new TSMVisualVP(_tsm, argv[i+1]);
    }
  }  
  printf("Failed to find the name for vertex shader (attributes version).\n" 
         "Usage: -vsa vertex_shader_name .\n");
  return NULL;
}

struct RegisterVPVisCreator
{
  RegisterVPVisCreator()
  { VisualizationFactory::connect('2', create_VP_vis); }
};

RegisterVPVisCreator register_vp_vis_creator__;

}

TSMVisualVP::TSMVisualVP(const TwoScaleModel& _tsm, const std::string& _vs_file_name)
: tsm_(_tsm), vs_file_name_(_vs_file_name), vertex_program_(NULL)
{}

TSMVisualVP::~TSMVisualVP()
{
  deactivate();
  unload_vertex_program();
  //clear constant data
  delete_VBOs(vbo_attribs_);
  delete_VBO(vbo_indices_);
#ifdef SEPARATE_PREC_ATTRIBS
  delete_VBO(vbo_vertices_);
  delete_VBO(vbo_prec_pt_);
  delete_VBO(vbo_prec_du_);
  delete_VBO(vbo_prec_dv_);
#else
  delete_VBO(vbo_fsv_blocks_);
#endif
}

void TSMVisualVP::unload_vertex_program()
{
  if (vertex_program_ != NULL)
  {
    vertex_program_->stop();
    delete vertex_program_;      
    vertex_program_ = NULL;
  }
}


/// IVisualization API

#ifndef SEPARATE_PREC_ATTRIBS
struct FSVBlock
{
  Vec4 vt_;
  Vec4 pt_;
  Vec4 du_;
  Vec4 dv_;
};
#endif//SEPARATE_PREC_ATTRIBS

bool TSMVisualVP::init()
{
  uint n_fs_vertices = tsm_.fs_vertices_.size();
  std::vector<Vec3> foot_pt(n_fs_vertices), foot_du(n_fs_vertices), foot_dv(n_fs_vertices);
  TwoScaleModel::CVCoeffArray cv_coeff;
  uint n_modified_cvs = tsm_.calc_geometry_split(&foot_pt.front(), &foot_du.front(), &foot_dv.front(), cv_coeff);
  
  int max_vertex_attribs;
  OGL_CHECK(::glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB, &max_vertex_attribs));
  if (max_vertex_attribs < (int)(n_modified_cvs + 5))
  {//vertices, color(s), f, du and dv precomputed fractions
    printf("ERROR: The number of required attributes (%d) is larger than "
           "the number of available vertex attributes (%d)!\n", n_modified_cvs + 5, max_vertex_attribs);
    return false;    
  }

  int max_vertex_uniform_components;
  OGL_CHECK(::glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB, &max_vertex_uniform_components));
  if (max_vertex_uniform_components < 3*(int)n_modified_cvs)
  {
    printf("ERROR: The number of modified control fsv_blocks components (%d) is larger than "
           "the number of available uniform vertex componenets(%d)!\n", 
           3*n_modified_cvs, max_vertex_uniform_components);
    return false;    
  }  
#ifdef SEPARATE_PREC_ATTRIBS
  std::vector<Vec4> vertices(n_fs_vertices), prec_pts(n_fs_vertices), 
                    prec_dus(n_fs_vertices), prec_dvs(n_fs_vertices);
  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    vertices[i][0] = tsm_.fs_vertices_[i].t0_displ_n_;
    vertices[i][1] = tsm_.fs_vertices_[i].t0_displ_du_;
    vertices[i][2] = tsm_.fs_vertices_[i].t0_displ_dv_;
    vertices[i][3] = tsm_.fs_vertices_[i].pt_displ_;

    prec_pts[i].v() = foot_pt[i];
    prec_pts[i][3] = tsm_.fs_vertices_[i].t1_displ_n_;
    
    prec_dus[i].v() = foot_du[i];
    prec_dus[i][3] = tsm_.fs_vertices_[i].t1_displ_du_;
    
    prec_dvs[i].v() = foot_dv[i];
    prec_dvs[i][3] = tsm_.fs_vertices_[i].t1_displ_dv_;
  }
  vbo_vertices_ = new_VBO(vertices);
  vbo_prec_pt_ = new_VBO(prec_pts);
  vbo_prec_du_ = new_VBO(prec_dus);
  vbo_prec_dv_ = new_VBO(prec_dvs);
#else      
  std::vector<FSVBlock> fsv_blocks(n_fs_vertices);
  for (uint i = 0; i < n_fs_vertices; ++i)
  {
    fsv_blocks[i].vt_[0] = tsm_.fs_vertices_[i].t0_displ_n_;
    fsv_blocks[i].vt_[1] = tsm_.fs_vertices_[i].t0_displ_du_;
    fsv_blocks[i].vt_[2] = tsm_.fs_vertices_[i].t0_displ_dv_;
    fsv_blocks[i].vt_[3] = tsm_.fs_vertices_[i].pt_displ_;

    fsv_blocks[i].pt_.v() = foot_pt[i];
    fsv_blocks[i].pt_[3] = tsm_.fs_vertices_[i].t1_displ_n_;
    
    fsv_blocks[i].du_.v() = foot_du[i];
    fsv_blocks[i].du_[3] = tsm_.fs_vertices_[i].t1_displ_du_;
    
    fsv_blocks[i].dv_.v() = foot_dv[i];
    fsv_blocks[i].dv_[3] = tsm_.fs_vertices_[i].t1_displ_dv_;
  }
  vbo_fsv_blocks_ = new_VBO(fsv_blocks);    
#endif//SEPARATE_PREC_ATTRIBS

  n_indices_ = tsm_.fs_tris_.size()*3;
  vbo_indices_ = new_VBO(tsm_.fs_tris_, GL_ELEMENT_ARRAY_BUFFER_ARB);
  
  //static data dependent of the number of modified control vertices
  vbo_attribs_.clear();
  vbo_attribs_.reserve(n_modified_cvs);
  
  modified_cvs_.clear(); 
  modified_cvs_.reserve(n_modified_cvs);
  
  for (uint i = 0, n = cv_coeff.size(); i < n; ++i)
  {
    if (!cv_coeff[i].empty())
    {
      modified_cvs_.push_back(i);
      vbo_attribs_.push_back(new_VBO(cv_coeff[i]));
    }
  }
  
  CHECK(modified_cvs_.size() == n_modified_cvs && vbo_attribs_.size() == n_modified_cvs);

  double total_static_mem = (vbo_attribs_.size()*sizeof(Vec3) + 4*sizeof(Vec4))*n_fs_vertices;
  total_static_mem += n_indices_*sizeof(uint);
  total_static_mem /= (1024.0*1024.0);  
  
  printf("Total static memory used %.5fMb\n", total_static_mem);
  //setup program
  if (!load_vertex_program())
  {
    return false;
  }
  activate();
  return true;
}

#define LINE_SIZE  1024

bool TSMVisualVP::load_vertex_program()
{
  if (vs_file_name_.empty())
  {
    printf("ERROR: No vertex shader file name specified.\n");
    return false;
  }
  
  unload_vertex_program();
  
  vertex_program_ = new VertexProgram();
  
  printf("Loading vertex shader source: %s ...\n", vs_file_name_.c_str());
  if (!vertex_program_->load_source(vs_file_name_.c_str()))
  {
    printf("ERROR: Loading of vertex shader source file: %s failed!\n", vs_file_name_.c_str());
    delete vertex_program_;
    vertex_program_ = NULL;
    return false;
  }
  std::vector<std::string> mod_lines, attrib_names, uniform_names;

  char buf[LINE_SIZE];
  mod_lines.push_back("/* attributes and uniforms of the modified control fsv_blocks */");
  for (uint i = 0, n = modified_cvs_.size(); i < n; ++i)
  {
    snprintf(buf, LINE_SIZE, "uniform vec3 cv%d;", i);
    mod_lines.push_back(buf);
    snprintf(buf, LINE_SIZE, "attribute vec3 cv_coeff%d;", i);
    mod_lines.push_back(buf);
    
    snprintf(buf, LINE_SIZE, "cv%d", i);
    uniform_names.push_back(buf);
    snprintf(buf, LINE_SIZE, "cv_coeff%d", i);
    attrib_names.push_back(buf);
  }  
  mod_lines.push_back("/* computes the \"foot\" frame for the current vertex */");
  mod_lines.push_back("void calc_foot_frame_dynamic(inout vec3 foot_pt, inout vec3 foot_du, inout vec3 foot_dv)");
  mod_lines.push_back("{");
  
  mod_lines.push_back("  /* add what is dependent on the modified CVs */");
  mod_lines.push_back("");
  for (uint i = 0, n = modified_cvs_.size(); i < n; ++i)
  {
    snprintf(buf, LINE_SIZE, "  foot_pt += cv_coeff%d.x * cv%d;", i, i); 
    mod_lines.push_back(buf);
    snprintf(buf, LINE_SIZE, "  foot_du += cv_coeff%d.y * cv%d;", i, i); 
    mod_lines.push_back(buf);
    snprintf(buf, LINE_SIZE, "  foot_dv += cv_coeff%d.z * cv%d;", i, i); 
    mod_lines.push_back(buf);
    mod_lines.push_back("");
  }
  mod_lines.push_back("}");
#if defined(CUSTOM_PREC_ATTRIBS) || defined(SEPARATE_PREC_ATTRIBS)
  mod_lines.push_back("");//define CUSTOM_PREC_ATTRIBS in the VS as well
  mod_lines.push_back("#define CUSTOM_PREC_ATTRIBS");
#endif
  vertex_program_->push_front_source(mod_lines); 
  if (!vertex_program_->compile_and_link())
  {
    printf("ERROR: Compilation/linking of the GPU program failed! Dumping source ..\n");
    dump();
    delete vertex_program_;
    vertex_program_ = NULL;
    return false;
  }
  vertex_program_->query_uniform_locations(uniform_names, loc_cvs_);
  vertex_program_->query_attrib_locations(attrib_names, loc_attribs_);

  loc_prec_foot_pt_ = vertex_program_->query_attrib_location("prec_foot_pt");
  loc_prec_foot_du_ = vertex_program_->query_attrib_location("prec_foot_du");
  loc_prec_foot_dv_ = vertex_program_->query_attrib_location("prec_foot_dv");
  return true;
}

void TSMVisualVP::activate()
{
  CHECK(vertex_program_);
  vertex_program_->start();  
  
#ifdef SEPARATE_PREC_ATTRIBS
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_vertices_));
  OGL_CHECK(::glEnableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glVertexPointer(4, GL_FLOAT, 0, (char*) NULL));
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_prec_pt_));
  OGL_CHECK(::glEnableVertexAttribArrayARB(loc_prec_foot_pt_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_pt_, 4, GL_FLOAT, 0, GL_FALSE, (char*) NULL));
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_prec_du_));
  OGL_CHECK(::glEnableVertexAttribArrayARB(loc_prec_foot_du_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_du_, 4, GL_FLOAT, 0, GL_FALSE, (char*) NULL));

  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_prec_dv_));
  OGL_CHECK(::glEnableVertexAttribArrayARB(loc_prec_foot_dv_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_dv_, 4, GL_FLOAT, 0, GL_FALSE, (char*) NULL));
#else//SEPARATE_PREC_ATTRIBS
  
  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_fsv_blocks_));

#ifdef CUSTOM_PREC_ATTRIBS
  OGL_CHECK(::glEnableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glVertexPointer(4, GL_FLOAT, sizeof(FSVBlock), (char *) NULL));
/*OGL_CHECK(::glEnableVertexAttribArrayARB(0));
  OGL_CHECK(::glVertexAttribPointerARB(0, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE, 
                                      (char *) NULL));*/  
  OGL_CHECK(::glEnableVertexAttribArrayARB(loc_prec_foot_pt_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_pt_, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE, 
                                      (char *) NULL + sizeof(Vec4)));
  
  OGL_CHECK(::glEnableVertexAttribArrayARB(loc_prec_foot_du_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_du_, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE,
                                      (char *) NULL + 2*sizeof(Vec4)));

  OGL_CHECK(::glEnableVertexAttribArrayARB(loc_prec_foot_dv_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_dv_, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE,
                                      (char *) NULL + 3*sizeof(Vec4)));
/*OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_pt_, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE, 
                                       &fsv_blocks[0].pt_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_du_, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE,
                                       &fsv_blocks[0].du_));
  OGL_CHECK(::glVertexAttribPointerARB(loc_prec_foot_dv_, 4, GL_FLOAT, sizeof(FSVBlock), GL_FALSE,
                                       &fsv_blocks[0].dv_));*/
#else
  OGL_CHECK(::glEnableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glVertexPointer(4, GL_FLOAT, sizeof(FSVBlock), (char *) NULL));
  
  OGL_CHECK(::glClientActiveTexture(GL_TEXTURE0));
  OGL_CHECK(::glEnableClientState(GL_TEXTURE_COORD_ARRAY));
  OGL_CHECK(::glTexCoordPointer(4, GL_FLOAT, sizeof(FSVBlock),
                                (char*)NULL + sizeof(Vec4)/*&fsv_blocks[0].pt_*/));

  OGL_CHECK(::glClientActiveTexture(GL_TEXTURE1));
  OGL_CHECK(::glEnableClientState(GL_TEXTURE_COORD_ARRAY));
  OGL_CHECK(::glTexCoordPointer(4, GL_FLOAT, sizeof(FSVBlock),
                                (char*) NULL + 2*sizeof(Vec4)/*&fsv_blocks[0].du_*/));

  OGL_CHECK(::glClientActiveTexture(GL_TEXTURE2));
  OGL_CHECK(::glEnableClientState(GL_TEXTURE_COORD_ARRAY));
  OGL_CHECK(::glTexCoordPointer(4, GL_FLOAT, sizeof(FSVBlock), 
                                (char*) NULL + 3*sizeof(Vec4)/*&fsv_blocks[0].dv_*/));
#endif

#endif//SEPARATE_PREC_ATTRIBS
  for (uint i = 0, n = loc_attribs_.size(); i < n; ++i)
  {//bind and set coefficient arrays' vbos
    OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_attribs_[i]));
    OGL_CHECK(::glEnableVertexAttribArrayARB(loc_attribs_[i]));
    OGL_CHECK(::glVertexAttribPointerARB(loc_attribs_[i], 3, GL_FLOAT, 0, GL_FALSE, (char *) NULL));
  }

  OGL_CHECK(::glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0));//unbind the array buffer
}

void TSMVisualVP::deactivate()
{
  if (vertex_program_ == NULL)
  {
    return;
  }
  vertex_program_->stop();

#ifdef CUSTOM_PREC_ATTRIBS
  OGL_CHECK(::glDisableClientState(GL_VERTEX_ARRAY));
  OGL_CHECK(::glDisableVertexAttribArrayARB(loc_prec_foot_pt_));
  OGL_CHECK(::glDisableVertexAttribArrayARB(loc_prec_foot_du_));
  OGL_CHECK(::glDisableVertexAttribArrayARB(loc_prec_foot_dv_));
#else
  OGL_CHECK(::glDisableClientState(GL_VERTEX_ARRAY));

  OGL_CHECK(::glClientActiveTexture(GL_TEXTURE0));
  OGL_CHECK(::glDisableClientState(GL_TEXTURE_COORD_ARRAY));

  OGL_CHECK(::glClientActiveTexture(GL_TEXTURE1));
  OGL_CHECK(::glDisableClientState(GL_TEXTURE_COORD_ARRAY));

  OGL_CHECK(::glClientActiveTexture(GL_TEXTURE2));
  OGL_CHECK(::glDisableClientState(GL_TEXTURE_COORD_ARRAY));
#endif

  for (uint i = 0, n = loc_attribs_.size(); i < n; ++i)
  {//disable coefficient arrays
    OGL_CHECK(::glDisableVertexAttribArrayARB(loc_attribs_[i]));
  }
}

void TSMVisualVP::draw(ModeEnum _mode)
{
/* the control fsv_blocks positions which we upload to the GPU here are the sole information
    which is transfered to the GPU at every frame */
  CHECK(loc_cvs_.size() == modified_cvs_.size());
  for (uint i = 0, n = loc_cvs_.size(); i < n; ++i)
  {
    const Vec3& cv = tsm_.cm_vertices_[modified_cvs_[i]];
    OGL_CHECK(::glUniform3fARB(loc_cvs_[i], cv.x, cv.y, cv.z));
  }
  
  //draw now
  if (_mode == MODE_POINTS)
  {
    OGL_CHECK(::glDrawArrays(GL_POINTS, 0, tsm_.fs_vertices_.size()));
  }
  else if (tsm_.prim_groups_ != NULL)
  {
    static uint ogl_pt[PT_FAN+1] = {GL_TRIANGLES, GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN};
    for (uint i = 0; i < tsm_.n_prim_groups_; ++i)
    {
      const PrimitiveGroup& pg = tsm_.prim_groups_[i];
      OGL_CHECK(::glDrawElements(ogl_pt[pg.type], pg.numIndices, GL_UNSIGNED_INT, pg.indices));
    }
  }
  else
  {
    //NOTE: indices from the sys. memory turned out to be faster - very surpising....!?
//     OGL_CHECK(::glDrawElements(GL_TRIANGLES, n_indices_, GL_UNSIGNED_INT, &tsm_.fs_tris_[0]));
    OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, vbo_indices_));
    OGL_CHECK(::glDrawElements(GL_TRIANGLES, n_indices_, GL_UNSIGNED_INT, (char *) NULL));
    OGL_CHECK(::glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0));//unbind indicies buffer
  }
}

void TSMVisualVP::dump()
{ 
  if (vertex_program_ == NULL)
  {
    printf("ERROR: vertex program not loaded.\n");
  }
  else
  {
    vertex_program_->dump_source(); 
  }
}

