/******************************************************************************
                       C++ Implementation: VertexProgram
                            -------------------
    description       :
    begin             : Wed Mar 23 2005
    author            : Martin Marinov, (C) 2005
    email             : marinov@cs.rwth-aachen.de
    copyright         : See LICENSE.txt file that comes with this distribution
 ******************************************************************************/

#include "VertexProgram.h"

#include <fstream>
#include <iostream>
#include <vector>
#include <string>

#define MAX_LINE_SIZE 4096

VertexProgram::VertexProgram()
: vsh_hnd_(0), program_hnd_(0), active_(false)
{}

VertexProgram::~VertexProgram()
{
  CHECK(!active_);
  if (program_hnd_ > 0)
  {
    OGL_CHECK(::glDeleteObjectARB(program_hnd_));
  }
  if (vsh_hnd_ > 0)
  {
    OGL_CHECK(::glDeleteObjectARB(vsh_hnd_));
  }
}

void print_info_log(GLhandleARB _hnd, const char* _msg = NULL)
{
  int log_size;
  OGL_CHECK(::glGetObjectParameterivARB(_hnd, GL_OBJECT_INFO_LOG_LENGTH_ARB, &log_size));
  if (log_size > 1)
  {
    char* log_buf = new char [log_size];
    int log_chars;
    OGL_CHECK(::glGetInfoLogARB(_hnd, log_size, &log_chars, log_buf));
    if (_msg != NULL)
    {
      printf("%s%s\n\n", _msg, log_buf);
    }
    else
    {
      printf("%s\n\n", log_buf);
    }
    delete [] log_buf;
  }
}

bool VertexProgram::load_source(const char* _file_name)
{
  std::ifstream ifs(_file_name);
  if (!ifs.is_open())
  {
    return false;
  }
  source_.clear();

  char buf[MAX_LINE_SIZE];
  while (!ifs.eof())
  {
    ifs.getline(buf, MAX_LINE_SIZE);
    source_ += buf;
  /*NOTE: the nvidia parser gets messed up sometimes if there is no '\n' at the end of the lines
    the whole source is not in a single string and does not recognizes #defines and C++ comments */
    source_ += " \n";
  }
  return true;
}

void VertexProgram::push_front_source(const std::vector<std::string>& _lines)
{
  std::string source_front;
  for (uint i = 0, n = _lines.size(); i < n; ++i)
  {
    source_front += _lines[i];
    source_front += " \n";
  }
  source_front += source_;
  source_ = source_front;
}

void VertexProgram::set_bindings(const StringArray& _bindings)
{ bindings_ = _bindings; }

bool VertexProgram::compile_and_link()
{
  OGL_CHECK(vsh_hnd_ = ::glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB));
  const char* source_str = source_.c_str();
  OGL_CHECK(::glShaderSourceARB(vsh_hnd_, 1, (const char**)&source_str, NULL));
  OGL_CHECK(::glCompileShaderARB(vsh_hnd_));
  int vs_compiled;
  OGL_CHECK(::glGetObjectParameterivARB(vsh_hnd_, GL_OBJECT_COMPILE_STATUS_ARB, &vs_compiled));   
  if (vs_compiled == 0)
  {
    printf("GLSLang vertex shader failed compilation.\n");
  }
  OGL_CHECK(program_hnd_ = ::glCreateProgramObjectARB());
  OGL_CHECK(::glAttachObjectARB(program_hnd_, vsh_hnd_));
  
  for (uint i = 0, n = bindings_.size(); i < n; ++i)
  {
    OGL_CHECK(::glBindAttribLocationARB(program_hnd_, i + 1, bindings_[i].c_str()));
  }
  
  OGL_CHECK(::glLinkProgramARB(program_hnd_));
  int prog_linked;
  OGL_CHECK(::glGetObjectParameterivARB(program_hnd_, GL_OBJECT_LINK_STATUS_ARB, &prog_linked));
  if (prog_linked == 0)
  {
    printf("GLSLang program failed linkage.\n");
  }
  print_info_log(vsh_hnd_, "GLSLang vertex shader log: \n");
  print_info_log(program_hnd_, "GLSLang program log: \n");
  return  prog_linked == 1 && vs_compiled == 1;
}


void VertexProgram::query_attrib_locations(const StringArray& _names, IndexArray& _locations) const
{
  _locations.resize(_names.size());
  for (uint i = 0, n = _names.size(); i < n; ++i)
  {
    OGL_CHECK(_locations[i] = query_attrib_location(_names[i]));
  }
}

int VertexProgram::query_attrib_location(const std::string& _name) const
{
  OGL_CHECK(int loc = ::glGetAttribLocationARB(program_hnd_, _name.c_str()));
  return loc;
}

void VertexProgram::query_uniform_locations(const StringArray& _names, IndexArray& _locations) const
{
  _locations.resize(_names.size());
  for (uint i = 0, n = _names.size(); i < n; ++i)
  {
    _locations[i] = query_uniform_location(_names[i].c_str());
  }
}

int VertexProgram::query_uniform_location(const std::string& _name) const
{
  OGL_CHECK(int loc = ::glGetUniformLocationARB(program_hnd_, _name.c_str()));
  return loc;
}

void VertexProgram::dump_source() const
{
  std::cout << source_;
/*  for (uint i = 0, n = source_.size(); i < n; ++i)
  {
    std::cout << source_[i] << std::endl;
  }
  std::cout << std::endl;
*/
}

